@@ -101,11 +101,12 @@ import Codec.Archive.LibZip.Errors
 import Data.Time.Clock (UTCTime, getCurrentTime)
 import Data.Time.Clock.POSIX (utcTimeToPOSIXSeconds)
 import Data.Word (Word8)
+import Control.Monad (when)
 import Control.Monad.State.Strict
     (StateT(..), MonadState(..), MonadTrans(..), lift, liftM)
 import Foreign.C.Error (Errno(..), eINVAL)
 import Foreign.C.String (withCString, withCStringLen, peekCString)
-import Foreign.C.Types (CInt, CSize)
+import Foreign.C.Types (CInt, CULLong)
 import Foreign.Marshal.Alloc (alloca)
 import Foreign.Marshal.Array (allocaArray, peekArray, withArrayLen, pokeArray)
 import Foreign.Marshal.Utils (with)
@@ -123,8 +124,8 @@ type Archive a = StateT Zip IO a
 -- | Monadic computation to read from open archive entries.
 -- See 'fromFile' and 'fromFileIx'.
 type Entry a = StateT
-    (ZipFile,Int,[FileFlag])   -- (file, position index, access flags)
-    (StateT Zip IO)            -- archive monad
+    (ZipFile,Integer,[FileFlag])   -- (file, position index, access flags)
+    (StateT Zip IO)                -- archive monad
     a
 
 --
@@ -152,14 +153,14 @@ withArchive flags path action =
         else return r
 
 -- | Get the number of entries in the archive.
-numFiles :: Archive Int
-numFiles = do
+numFiles :: [FileFlag] -> Archive Integer
+numFiles flags  = do
   z <- getZip
-  lift $ fromIntegral `liftM` c'zip_get_num_files z
+  lift $ fromIntegral `liftM` c'zip_get_num_entries z (combine flags)
 
 -- | Get name of an entry in the archive by its index.
 fileName :: [FileFlag]  -- ^ 'FileUNCHANGED' flag can be used.
-         -> Int         -- ^ Position index of a file in the archive.
+         -> Integer     -- ^ Position index of a file in the archive.
          -> Archive FilePath  -- ^ Name of the file in the archive.
 fileName flags i = do
   z <- getZip
@@ -170,7 +171,7 @@ fileName flags i = do
 -- | Locate an entry (get its index) in the archive by its name.
 nameLocate :: [FileFlag]  -- ^ Filename lookup mode.
            -> FilePath    -- ^ Name of the file in the archive.
-           -> Archive (Maybe Int)  -- ^ 'Just' position index if found.
+           -> Archive (Maybe Integer)  -- ^ 'Just' position index if found.
 nameLocate flags name = do
   z <- getZip
   lift $
@@ -184,19 +185,19 @@ nameLocate flags name = do
 fileNames :: [FileFlag]  -- ^ 'FileUNCHANGED' flag is accepted.
           -> Archive [FilePath]
 fileNames flags = do
-  n <- numFiles
+  n <- numFiles flags
   mapM (fileName flags) [0..n-1]
 
 -- | Get size of a file in the archive.
 fileSize :: [FileFlag]  -- ^ Filename lookup mode, 'FileUNCHANGED' can be used.
          -> FilePath    -- ^ Name of the file in the archive.
-         -> Archive Int -- ^ File size.
+         -> Archive Integer -- ^ File size.
 fileSize flags name = fileStat flags name >>= return . zs'size
 
 -- | Get size of a file in the archive (by index).
 fileSizeIx :: [FileFlag]  -- ^ 'FileUNCHANGED' is accepted.
-           -> Int         -- ^ Position index of a file in the archive.
-           -> Archive Int -- ^ File size.
+           -> Integer     -- ^ Position index of a file in the archive.
+           -> Archive Integer -- ^ File size.
 fileSizeIx flags i = fileStatIx flags i >>= return . zs'size
 
 -- | Get information about a file in the archive.
@@ -214,7 +215,7 @@ fileStat flags name = do
 
 -- | Get information about a file in the archive (by index).
 fileStatIx :: [FileFlag]  -- ^ 'FileUNCHANGED' can be used.
-           -> Int         -- ^ Position index of a file in the archive.
+           -> Integer     -- ^ Position index of a file in the archive.
            -> Archive ZipStat  -- ^ Information about the file.
 fileStatIx flags i = do
   z <- getZip
@@ -232,7 +233,7 @@ deleteFile flags name = do
   maybe (lift $ E.throwIO ErrNOENT) deleteFileIx mbi
 
 -- | Delete file (referenced by position index) from the archive.
-deleteFileIx :: Int  -- ^ Position index of a file in the archive.
+deleteFileIx :: Integer  -- ^ Position index of a file in the archive.
              -> Archive ()
 deleteFileIx i = do
   z <- getZip
@@ -251,7 +252,7 @@ renameFile flags oldname newname = do
   maybe (lift $ E.throwIO ErrNOENT) (\i -> renameFileIx i newname) mbi
 
 -- | Rename file (referenced by position index) in the archive.
-renameFileIx :: Int  -- ^ Position index of a file in the archive.
+renameFileIx :: Integer  -- ^ Position index of a file in the archive.
              -> FilePath -- ^ New name.
              -> Archive ()
 renameFileIx i newname = do
@@ -294,7 +295,7 @@ replaceFile flags name src = do
         (\i -> replaceFileIx i src >> return ()) mbi
 
 -- | Replace a file in the archive (referenced by position index).
-replaceFileIx :: Int       -- ^ Position index of a file in the archive.
+replaceFileIx :: Integer   -- ^ Position index of a file in the archive.
               -> ZipSource -- ^ Source where the new file data is obtained from
               -> Archive ()
 replaceFileIx i src = do
@@ -320,8 +321,8 @@ sourceBuffer src = do
 
 -- | Create a data source from a file.
 sourceFile :: FilePath   -- ^ File to open.
-           -> Int        -- ^ Offset from the beginning of the file.
-           -> Int        -- ^ The number of bytes to read. If @0@ or @-1@,
+           -> Integer    -- ^ Offset from the beginning of the file.
+           -> Integer    -- ^ The number of bytes to read. If @0@ or @-1@,
                          -- the read till the end of file.
            -> Archive ZipSource
 sourceFile name offset len = do
@@ -335,9 +336,9 @@ sourceFile name offset len = do
 -- | Create a data source from a file in the zip archive.
 sourceZip :: [FileFlag]  -- ^ 'FileUNCHANGED' and 'FileRECOMPRESS' can be used.
           -> Zip         -- ^ Source archive.
-          -> Int         -- ^ Position index of a file in the source archive.
-          -> Int         -- ^ Offset from the beginning of the file.
-          -> Int         -- ^ The number of bytes to read. If @0@ or @-1@,
+          -> Integer     -- ^ Position index of a file in the source archive.
+          -> Integer     -- ^ Offset from the beginning of the file.
+          -> Integer     -- ^ The number of bytes to read. If @0@ or @-1@,
                          -- then read till the end of file.
           -> Archive ZipSource
 sourceZip flags srcz srcidx offset len = do
@@ -351,8 +352,8 @@ sourceZip flags srcz srcidx offset len = do
 
 -- | Create a data source from a 'PureSource'.
 -- Note: input of @[a]@ is converted to @[Word8]@ internally.
-sourcePure :: (Enum a, Storable a, Storable st)
-           => PureSource a st -> Archive ZipSource
+sourcePure :: (Enum a, Storable a, Storable st, Integral szt)
+           => PureSource a st szt -> Archive ZipSource
 sourcePure pureSrc = do
   z <- getZip
   lift $ do
@@ -366,18 +367,20 @@ sourcePure pureSrc = do
 -- | Wrapper for a user-provided pure function to be used with 'sourcePure'.
 -- Data size should be known in advance ('srcSize').
 -- The function should support reading by chunks ('readSrc').
-data (Enum a, Storable a, Storable st) => PureSource a st = PureSource {
-    srcState :: st   -- ^ Initial state of the source.
-  , srcSize  :: Int  -- ^ Total size of the data.
-  , srcMTime :: Maybe UTCTime  -- ^ Modification time (current time if Nothing).
-  , readSrc  :: Int -> st -> Maybe (Int, [a], st) -- ^ Read a chunk of the data,
-                -- return @Just@ the size of data read, the data themselves and
-                -- the new state of the source, or @Nothing@ on error.
+data PureSource a st szt = PureSource {
+    srcState :: st     -- ^ Initial state of the source.
+  , srcSize  :: szt    -- ^ Total size of the data.
+  , srcMTime :: Maybe UTCTime
+                -- ^ Modification time (current time if Nothing).
+  , readSrc  :: szt -> st -> Maybe (szt, [a], st)
+                -- ^ Read a chunk of the data, return @Just@ the size
+                -- of data read, the data themselves and the new state
+                -- of the source, or @Nothing@ on error.
   }
 
-runPureSource :: (Enum a, Storable a, Storable st)
-              => PureSource a st
-              -> (Ptr () -> Ptr () -> CSize -> C'zip_source_cmd -> IO CSize)
+runPureSource :: (Enum a, Storable a, Storable st, Integral szt) =>
+       PureSource a st szt
+    -> (Ptr () -> Ptr () -> CULLong -> C'zip_source_cmd -> IO CULLong)
 runPureSource src pState pData len cmd
   | cmd == c'ZIP_SOURCE_OPEN = return 0
   | cmd == c'ZIP_SOURCE_READ = do
@@ -451,7 +454,7 @@ getFileComment flags name = do
 
 -- | Get comment for a file in the archive (referenced by position index).
 getFileCommentIx :: [FileFlag]  -- ^ FileUNCHANGED can be used.
-                 -> Int         -- ^ Position index of the file.
+                 -> Integer     -- ^ Position index of the file.
                  -> Archive (Maybe String)
 getFileCommentIx flags i = do
   z <- getZip
@@ -473,7 +476,7 @@ setFileComment flags path comment = do
   maybe (lift $ E.throwIO ErrNOENT) (flip setFileCommentIx comment) mbi
 
 -- | Set comment for a file in the archive (referenced by position index).
-setFileCommentIx :: Int        -- ^ Position index of a file in the archive.
+setFileCommentIx :: Integer    -- ^ Position index of a file in the archive.
                  -> String     -- ^ New file comment. 
                  -> Archive ()
 setFileCommentIx i comment = do
@@ -493,7 +496,7 @@ removeFileComment flags path = do
   maybe (lift $ E.throwIO ErrNOENT) removeFileCommentIx mbi
 
 -- | Remove comment for a file in the archive (referenced by position index).
-removeFileCommentIx :: Int  -- ^ Position index of a file in the archive.
+removeFileCommentIx :: Integer -- ^ Position index of a file in the archive.
                     -> Archive ()
 removeFileCommentIx i = do
   z <- getZip
@@ -511,7 +514,7 @@ unchangeFile flags name = do
   maybe (lift $ E.throw ErrNOENT) unchangeFileIx mbi
 
 -- | Undo changes to a file in the archive (referenced by position index).
-unchangeFileIx :: Int  -- ^ Position index of a file in the archive.
+unchangeFileIx :: Integer  -- ^ Position index of a file in the archive.
                -> Archive ()
 unchangeFileIx i = do
   z <- getZip
@@ -576,7 +579,7 @@ fromFile flags name action = do
 -- an 'Archive' action (see also 'withArchive'). 'fromFileIx' can be replaced
 -- with 'fileContentsIx' to read an entire file at once.
 fromFileIx :: [FileFlag] -- ^ 'FileCOMPRESSED' and 'FileUNCHANGED' can be used.
-           -> Int        -- ^ Position index of a file in the archive.
+           -> Integer    -- ^ Position index of a file in the archive.
            -> Entry a    -- ^ Action with the file.
            -> Archive a
 fromFileIx flags i action = do
@@ -594,11 +597,14 @@ fromFileIx flags i action = do
 -- | Read at most @n@ bytes from the file.
 readBytes ::
     (Enum a)
-    => Int       -- ^ The number of bytes to read.
+    => Integer   -- ^ The number of bytes to read.
     -> Entry [a] -- ^ Bytes read.
 readBytes n = do
+  lift . lift $
+    when (n > toInteger (maxBound::Int))
+      (E.throwIO ErrMEMORY) -- allocaArray can't allocate > (maxBound::Int)
   (zf,_,_) <- get
-  lift . lift $ allocaArray n $ \buf -> do
+  lift . lift $ allocaArray (fromIntegral n) $ \buf -> do
          nread <- c'zip_fread zf (castPtr buf) (fromIntegral n)
          if nread < 0
             then
@@ -608,7 +614,7 @@ readBytes n = do
               return . map (toEnum . fromEnum) $ bs
 
 -- | Skip @n@ bytes from the open file. Note: this is not faster than reading.
-skipBytes :: Int -> Entry ()
+skipBytes :: Integer -> Entry ()
 skipBytes n = (readBytes n :: Entry [Word8]) >> return ()
 
 -- | Read entire file contents.
@@ -631,7 +637,7 @@ fileContents flags name = fromFile flags name readContents
 -- 'readContents' from within 'Archive' monad.
 fileContentsIx :: (Enum a)
     => [FileFlag]
-    -> Int
+    -> Integer
     -> Archive [a]
 fileContentsIx flags i = fromFileIx flags i readContents
 
